(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Strengthen functions into simpler monads.
 *
 * Each block of lifting lemmas converts functions in the "L2" monadic
 * framework (an exception framework) into its own framework.
 *)

theory TypeStrengthen
imports
  L2Defs
  "../lib/OptionMonadND"
  ExecConcrete
begin

(*
 * Helpers for exception polymorphism lemmas (L2_call_Foo_polymorphic).
 *
 * They are used to rewrite a term like
 *
 *   L2_call x = Foo y
 *
 * into an identical term with a different exception type.
 *)

definition
  unliftE
where
  "unliftE (x :: ('a, 'u + 'b) nondet_monad) \<equiv> x <catch> (\<lambda>_. fail)"

lemma L2_call_liftE_unliftE:
    "L2_call M x = liftE (unliftE (L2_call M x))"
  apply (clarsimp simp: L2_call_def unliftE_def)
  apply (rule ext)
  apply (clarsimp simp: handleE'_def catch_def liftE_def bind_assoc)
  apply (clarsimp cong: bind_apply_cong)
  apply (clarsimp simp: bind_def split_def return_def split: sum.splits)
  apply (force simp: return_def fail_def split: sum.splits)+
  done

lemma unliftE_liftE [simp]:
    "unliftE (liftE x) = x"
  apply (clarsimp simp: unliftE_def catch_liftE)
  done

(* Lift recursive calls into measured calls. Used by type_strengthen.ML. *)
lemma lift_recursive_call_proof:
  "(\<And>measure'. L2_call (\<lambda>_. measure') hl = ts measure')
    \<Longrightarrow> L2_call M hl = L2_seq (L2_gets M [''measure'']) ts"
  apply (monad_eq simp: L2_call_def L2_defs)
  done



(*
 * Lifting into pure functional Isabelle.
 *)

definition "TS_return x \<equiv> liftE (return x)"

lemma L2_call_TS_return: "L2_call M (\<lambda>m. TS_return (a m)) = L2_gets (\<lambda>s. a (M s)) [''ret'']"
  apply (monad_eq simp: L2_call_def L2_gets_def TS_return_def)
  done

lemma TS_return_L2_gets:
    "L2_gets (\<lambda>_. P) n = TS_return P"
  by (monad_eq simp: L2_defs TS_return_def)

lemma L2_call_L2_gets_polymorphic:
  "(L2_call M x :: ('s, 'a, 'e1) L2_monad) = L2_gets y n
    \<Longrightarrow> (L2_call M x  :: ('s, 'a, 'e2) L2_monad) = L2_gets y n"
  apply (monad_eq simp: L2_defs L2_call_def Ball_def split: sum.splits)
  apply blast
  done

setup {*
Monad_Types.new_monad_type
  "pure"
  "Pure function"
  (Monad_Types.check_lifting_head [@{term "TS_return"}])
  100
  @{thms L2_call_TS_return TS_return_L2_gets}
  @{term "(%a. L2_gets (%_. a) [''ret'']) :: 'a => ('s, 'a, unit) L2_monad"}
  @{thm L2_call_L2_gets_polymorphic}
  #2
|> Context.theory_map
*}

lemma TS_return_L2_seq:
    "L2_seq (TS_return A) (\<lambda>a. TS_return (B a))
       = TS_return (let a = A in B a)"
  by (monad_eq simp: L2_defs TS_return_def)

lemma TS_return_L2_condition:
    "L2_condition (\<lambda>_. c) (TS_return A) (TS_return B) = TS_return (if c then A else B)"
  by (monad_eq simp: L2_defs TS_return_def)

lemmas [ts_rule pure] =
  TS_return_L2_gets
  TS_return_L2_seq
  TS_return_L2_condition
  split_distrib[where T = TS_return]

lemma L2_seq_TS_return:
    "TS_return (let a = A in B a) = L2_seq (L2_gets (\<lambda>_. A) []) (\<lambda>a. L2_gets (\<lambda>_. B a) [])"
  by (monad_eq simp: L2_defs TS_return_def)

lemma L2_condition_TS_return:
    "TS_return (if c then A else B) = L2_condition (\<lambda>_. c) (L2_gets (\<lambda>_. A) []) (L2_gets (\<lambda>_. B) [])"
  by (monad_eq simp: L2_defs TS_return_def)

lemmas [ts_rule pure unlift] =
  TS_return_L2_gets      [where n = "[]", symmetric]
  TS_return_L2_seq       [symmetric]
  TS_return_L2_condition [symmetric]
  L2_seq_TS_return
  L2_condition_TS_return
  split_distrib[where T = TS_return, symmetric]



(*
 * Lifting into pure functional Isabelle with state.
 *)

definition "TS_gets x \<equiv> liftE (gets x)"

lemma TS_gets_L2_gets:
    "L2_gets X n = TS_gets X"
  by (monad_eq simp: L2_defs TS_gets_def)

lemma L2_call_TS_gets: "L2_call M (\<lambda>m. TS_gets (a m)) = L2_gets (\<lambda>s. a (M s) s) [''TS_internal_retval'']"
  apply (monad_eq simp: L2_call_def L2_gets_def TS_gets_def)
  done

setup {*
Monad_Types.new_monad_type
  "gets"
  "Read-only function"
  (Monad_Types.check_lifting_head [@{term "TS_gets"}])
  80
  @{thms L2_call_TS_gets TS_gets_L2_gets}
  @{term "(%x. L2_gets x [''ret'']) :: ('s => 'a) => ('s, 'a, unit) L2_monad"}
  @{thm L2_call_L2_gets_polymorphic}
  (fn (state, ret, ex) => state --> ret)
|> Context.theory_map
*}

lemma TS_gets_L2_seq:
    "L2_seq (TS_gets A) (\<lambda>x. TS_gets (B x)) = (TS_gets (\<lambda>s. let x = A s in B x s))"
  by (monad_eq simp: L2_defs TS_gets_def)

lemma TS_gets_L2_condition:
    "L2_condition c (TS_gets A) (TS_gets B) = TS_gets (\<lambda>s. if c s then (A s) else (B s))"
  by (monad_eq simp: L2_defs TS_gets_def)

lemmas [ts_rule gets] =
  TS_gets_L2_gets
  TS_gets_L2_seq
  TS_gets_L2_condition
  split_distrib[where T = TS_gets]

lemmas [ts_rule gets unlift] =
  TS_gets_L2_gets      [where n = "[]", symmetric]
  TS_gets_L2_seq       [symmetric]
  TS_gets_L2_condition [symmetric]
  split_distrib[where T = TS_gets, symmetric]



(*
 * Lifting into option monad.
 *)

definition "gets_theE \<equiv> \<lambda>x. (liftE (gets_the x))"

lemma L2_call_gets_theE [simp]: "L2_call M (\<lambda>m. gets_theE (x m)) = (gets M >>= (\<lambda>m. gets_theE (x m)))"
  apply (monad_eq simp: L2_call_def L2_gets_def gets_theE_def gets_the_def gets_bind_ign exec_gets)
  done

lemma liftE_gets_theE: "gets_theE X = liftE (gets_the X)"
  apply (clarsimp simp: gets_theE_def)
  done
lemma L2_call_gets_theE_polymorphic:
  "(L2_call M x :: ('s, 'a, 'e1) L2_monad) = gets_theE y
    \<Longrightarrow> (L2_call M x  :: ('s, 'a, 'e2) L2_monad) = gets_theE y"
  apply (metis L2_call_liftE_unliftE liftE_gets_theE unliftE_liftE)
  done

lemma in_gets_theE [monad_eq]:
     "(rv, s') \<in> fst (gets_theE M s) = (\<exists>v'. rv = Inr v' \<and> s' = s \<and> M s = Some v')"
  apply (monad_eq simp: gets_theE_def)
  done

lemma snd_gets_theE [monad_eq]:
     "snd (gets_theE M s) = (M s = None)"
  apply (monad_eq simp: gets_theE_def gets_the_def Bex_def assert_opt_def split: option.splits)
  done

lemma gets_theE_ofail [simp]:
  "gets_theE ofail = fail"
  by (monad_eq simp: L2_defs ofail_def split: option.splits)

setup {*
Monad_Types.new_monad_type
  "option"
  "Option monad"
  (Monad_Types.check_lifting_head [@{term "gets_theE"}])
  60
  @{thms L2_call_gets_theE gets_theE_ofail}
  @{term "gets_theE :: ('s => 'a option) => ('s, 'a, unit) L2_monad"}
  @{thm L2_call_gets_theE_polymorphic}
  (fn (state, ret, ex) =>
     state --> Term.map_atyps (fn t => if t = @{typ "'a"} then ret else t) @{typ "'a option"})
|> Context.theory_map
*}

lemma gets_theE_L2_gets:
  "L2_gets a n = gets_theE (ogets a)"
  by (monad_eq simp: L2_defs ogets_def)

lemma gets_theE_L2_seq:
  "L2_seq (gets_theE X) (\<lambda>x. gets_theE (Y x)) = gets_theE (X |>> Y)"
  by (monad_eq simp: L2_defs ogets_def Bex_def obind_def split: option.splits)

lemma gets_theE_L2_guard:
  "L2_guard G = gets_theE (oguard G)"
  by (monad_eq simp: L2_defs oguard_def split: option.splits)

lemma gets_theE_L2_condition:
  "L2_condition C (gets_theE L) (gets_theE R) = gets_theE (ocondition C L R)"
  by (monad_eq simp: L2_defs ocondition_def split: option.splits)

lemma gets_theE_L2_fail:
  "L2_fail = gets_theE (ofail)"
  by (monad_eq simp: L2_defs ofail_def split: option.splits)

lemma gets_theE_L2_recguard:
  "L2_recguard m (gets_theE x) = gets_theE (ocondition (\<lambda>_. m = 0) ofail x)"
  by (monad_eq simp: L2_defs ocondition_def ofail_def split: option.splits)

lemma gets_theE_L2_while:
  "L2_while C (\<lambda>x. gets_theE (B x)) i n = gets_theE (owhile C B i)"
  unfolding L2_while_def gets_theE_def gets_the_whileLoop[symmetric]
  by (rule whileLoopE_liftE)

lemmas [ts_rule option] =
  gets_theE_L2_seq
  gets_theE_L2_fail
  gets_theE_L2_guard
  gets_theE_L2_recguard
  gets_theE_L2_gets
  gets_theE_L2_condition
  gets_theE_L2_while
  split_distrib[where T = gets_theE]

lemmas [ts_rule option unlift] =
  gets_theE_L2_seq       [symmetric]
  gets_theE_L2_fail      [symmetric]
  gets_theE_L2_guard     [symmetric]
  gets_theE_L2_recguard  [symmetric]
  gets_theE_L2_gets      [where n = "[]", symmetric]
  gets_theE_L2_condition [symmetric]
  gets_theE_L2_while     [symmetric]
  split_distrib[where T = gets_theE, symmetric]


(*
 * Lifting into the nondeterministic state monad.
 * All L2 terms can be lifted into it.
 *)

lemma L2_call_liftE_polymorphic:
  "((L2_call M x) :: ('s, 'a, 'e1) L2_monad) = liftE y
    \<Longrightarrow> (L2_call M x :: ('s, 'a, 'e2) L2_monad) = liftE y"
  apply (metis L2_call_liftE_unliftE unliftE_liftE)
  done

setup {*
Monad_Types.new_monad_type
  "nondet"
  "Nondeterministic state monad (default)"
  (Monad_Types.check_lifting_head [@{term "liftE"}])
  0
  @{thms L2_call_liftE}
  @{term "liftE :: ('s, 'a) nondet_monad => ('s, 'a, unit) L2_monad"}
  @{thm L2_call_liftE_polymorphic}
  (fn (state, ret, ex) =>
     Term.map_atyps (fn t => if t = @{typ "'a"} then ret
                               else if t = @{typ "'s"} then state else t)
       @{typ "('s, 'a) nondet_monad"})
|> Context.theory_map
*}

lemma liftE_L2_seq: "L2_seq (liftE A) (\<lambda>x. liftE (B x)) = (liftE (A >>= B))"
  apply (clarsimp simp: L2_defs)
  apply (clarsimp simp: liftE_def bindE_def bind_assoc)
  done

lemma liftE_L2_condition: "L2_condition c (liftE A) (liftE B) = liftE (condition c A B)"
  apply (clarsimp simp: L2_defs)
  apply (rule ext)+
  apply monad_eq
  apply blast
  done

lemma liftE_L2_modify: "L2_modify m = liftE (modify m)"
  apply (clarsimp simp: L2_defs)
  done

lemma liftE_L2_gets: "L2_gets a n = liftE (gets a)"
  apply (clarsimp simp: L2_defs)
  done

lemma liftE_L2_recguard:
  "(L2_recguard x (liftE A)) = liftE (condition (\<lambda>s. x > 0) A fail)"
  apply (case_tac "x = 0")
   apply clarsimp
  apply (clarsimp simp: L2_recguard_def)
  done

lemma liftE_L2_while: "L2_while c (\<lambda>r. liftE (B r)) i n = liftE (whileLoop c B i)"
  apply (clarsimp simp: L2_while_def)
  apply (rule whileLoopE_liftE)
  done

lemma liftE_L2_throw: "L2_throw X n = throwError X"
  apply (monad_eq simp: L2_throw_def)
  done

lemma liftE_L2_catch: "L2_catch (liftE A) B = liftE A"
  apply (clarsimp simp: L2_defs)
  done

lemma liftE_L2_catch': "L2_catch A (\<lambda>x. liftE (B x)) = liftE (A <catch> B)"
  apply (clarsimp simp: L2_defs)
  apply (clarsimp simp: handleE'_def liftE_def catch_def bind_assoc)
  apply (rule arg_cong [where f="\<lambda>x. (A >>= x)"])
  apply (rule ext)
  apply (clarsimp split: sum.splits)
  done

lemma liftE_L2_unknown: "L2_unknown name = liftE (select UNIV)"
  apply (clarsimp simp: L2_defs)
  done

lemma liftE_L2_spec: "L2_spec S = liftE (spec S >>= (\<lambda>_. select UNIV))"
  apply (clarsimp simp: L2_defs)
  done

lemma liftE_L2_guard: "L2_guard G = liftE (guard G)"
  apply (clarsimp simp: L2_defs)
  done

lemma liftE_L2_fail: "L2_fail = liftE (fail)"
  apply (clarsimp simp: L2_defs liftE_def)
  done

lemma liftE_exec_concrete:
  "exec_concrete st (liftE x) = liftE (exec_concrete st x)"
  apply (rule monad_eqI)
    apply (clarsimp simp: in_liftE in_exec_concrete)
    apply force
   apply (clarsimp simp: in_liftE in_exec_concrete)
   apply force
  apply (clarsimp simp: snd_exec_concrete snd_liftE)
  done

lemma liftE_exec_abstract:
  "exec_abstract st (liftE x) = liftE (exec_abstract st x)"
  apply (rule monad_eqI)
    apply (clarsimp simp: in_liftE in_exec_abstract)
   apply (clarsimp simp: in_liftE in_exec_abstract)
  apply (clarsimp simp: snd_exec_abstract snd_liftE)
  done

lemmas [ts_rule nondet] =
  liftE_L2_seq
  liftE_L2_condition
  liftE_L2_modify
  liftE_L2_gets
  liftE_L2_while
  liftE_L2_throw
  liftE_L2_catch
  liftE_L2_catch'
  liftE_L2_spec
  liftE_L2_guard
  liftE_L2_unknown
  liftE_L2_fail
  liftE_L2_recguard
  liftE_exec_concrete
  liftE_exec_abstract
  liftE_gets_theE
  split_distrib [where T=liftE]

lemmas [ts_rule nondet unlift] =
  liftE_L2_seq        [symmetric]
  liftE_L2_condition  [symmetric]
  liftE_L2_modify     [symmetric]
  liftE_L2_gets       [symmetric]
  liftE_L2_while      [symmetric]
  liftE_L2_throw      [symmetric]
  liftE_L2_catch      [symmetric]
  liftE_L2_catch'     [symmetric]
  liftE_L2_spec       [symmetric]
  liftE_L2_guard      [symmetric]
  liftE_L2_unknown    [symmetric]
  liftE_L2_fail       [symmetric]
  liftE_L2_recguard   [symmetric]
  liftE_exec_concrete [symmetric]
  liftE_exec_abstract [symmetric]
  liftE_gets_theE     [symmetric]
  split_distrib [where T=liftE, symmetric]

end
