(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Support for the top-level "autocorres" command.
 *)
structure AutoCorres =
struct

(*
 * Option parsing for the autocorres command.
 * The most general form of the command is
 *   autocorres [ no_heap_abs = FUNC_NAMES,
 *                ts_rules = RULE_NAMES,
 *                ts_force RULE_NAME = FUNC_NAMES,
 *                ts_force ... = ...,
 *                statistics
 *              ]
 *              "prog.c"
 *)

(* Most fields are wrapped in option so that the parser can work out
   whether they have been specified already.
   We don't allow an option to be specified multiple times. Perhaps we should. *)
type autocorres_options = {
  (* Do not lift heaps for these functions. *)
  no_heap_abs : string list option,

  (* Insist the the following functions should be lifted, even if our
   * heuristics claim it won't succeed. *)
  force_heap_abs : string list option,

  (* Functions to word abstract. *)
  word_abs : string list option,

  (* Only lift to these monads. *)
  ts_rules : string list option,

  (* Force functions to be lifted to certain monads.
     The symtab is keyed on function name. *)
  ts_force : string Symtab.table,

  print_stats : bool option
}

(* Boilerplate. *)
fun acopt_set_no_heap_abs funcs dup_err (opt : autocorres_options) : autocorres_options =
  case #no_heap_abs opt of
      NONE => { no_heap_abs = SOME funcs,
                force_heap_abs = #force_heap_abs opt,
                word_abs = #word_abs opt,
                ts_rules = #ts_rules opt,
                ts_force = #ts_force opt,
                print_stats = #print_stats opt }
    | SOME _ => error dup_err

fun acopt_set_force_heap_abs funcs dup_err (opt : autocorres_options) : autocorres_options =
  case #force_heap_abs opt of
      NONE => { no_heap_abs = #no_heap_abs opt,
                force_heap_abs = SOME funcs,
                word_abs = #word_abs opt,
                ts_rules = #ts_rules opt,
                ts_force = #ts_force opt,
                print_stats = #print_stats opt }
    | SOME _ => error dup_err

fun acopt_set_word_abs funcs dup_err (opt : autocorres_options) : autocorres_options =
  case #word_abs opt of
      NONE => { no_heap_abs = #no_heap_abs opt,
                force_heap_abs = #force_heap_abs opt,
                word_abs = SOME funcs,
                ts_rules = #ts_rules opt,
                ts_force = #ts_force opt,
                print_stats = #print_stats opt }
    | SOME _ => error dup_err

fun acopt_set_ts_rules rules dup_err (opt : autocorres_options) : autocorres_options =
  case #ts_rules opt of
      NONE => { no_heap_abs = #no_heap_abs opt,
                force_heap_abs = #force_heap_abs opt,
                word_abs = #word_abs opt,
                ts_rules = SOME rules,
                ts_force = #ts_force opt,
                print_stats = #print_stats opt }
    | SOME _ => error dup_err

(* dup_err : (func, old_rule) -> err_msg *)
fun acopt_set_ts_force rule funcs (dup_err : (string * string) -> string)
                       (opt : autocorres_options) : autocorres_options =
  let val ts_force' = fold (fn func => (fn table =>
                              Symtab.update_new (func, rule) table
                              handle Symtab.DUP _ =>
                                let val old_rule = the (Symtab.lookup table func) in
                                  if rule = old_rule then table else
                                  dup_err (func, old_rule) |> error
                                end))
                           funcs (#ts_force opt)
  in { no_heap_abs = #no_heap_abs opt,
       force_heap_abs = #force_heap_abs opt,
       word_abs = #word_abs opt,
       ts_rules = #ts_rules opt,
       ts_force = ts_force',
       print_stats = #print_stats opt } end

fun acopt_set_print_stats do_it dup_err (opt : autocorres_options) : autocorres_options =
  case #print_stats opt of
      NONE => { no_heap_abs = #no_heap_abs opt,
                force_heap_abs = #force_heap_abs opt,
                word_abs = #word_abs opt,
                ts_rules = #ts_rules opt,
                ts_force = #ts_force opt,
                print_stats = SOME do_it }
    | SOME _ => error dup_err


fun expect x y = !! (K (K ("autocorres: expected " ^ y ^ " after " ^ x)))

(* Generic parser for "NAME = STRING ..." *)
fun named_opt name elem_desc=
  Parse.reserved name |--
  expect (quote name) "\"=\"" (Parse.$$$ "=" |--
  expect "\"=\"" elem_desc (Scan.repeat Parse.text))

(* Valid options. *)
val no_heap_abs_parser =
  named_opt "no_heap_abs" "function names" >>
  (fn funcs => acopt_set_no_heap_abs funcs "autocorres: no_heap_abs option specified multiple times")

val force_heap_abs_parser =
  named_opt "force_heap_abs" "function names" >>
  (fn funcs => acopt_set_force_heap_abs funcs "autocorres: force_heap_abs option specified multiple times")

val ts_rules_parser =
  named_opt "ts_rules" "rule names" >>
  (fn rules => acopt_set_ts_rules rules "autocorres: ts_rules option specified multiple times")

val ts_force_parser =
  ((Parse.reserved "ts_force" |--
      expect "\"ts_force\"" "rule name"
      (Parse.text :-- (fn name => expect name "\"=\"" (Parse.$$$ "="))) --
    Scan.repeat Parse.text)) >>
  (fn ((rule, _), funcs) => acopt_set_ts_force rule funcs
                         (fn (func, old_rule) => "autocorres: function " ^ quote func ^
                            " is already being forced to type " ^ quote old_rule))

val word_abs_parser =
  named_opt "word_abs" "function names" >>
  (fn funcs => acopt_set_word_abs funcs "autocorres: word_abs option specified multiple times")

val print_stats_parser =
  Parse.reserved "statistics" >>
  (fn _ => acopt_set_print_stats true "autocorres: statistics option specified multiple times")

(* Combined parser. *)
val autocorres_parser : (autocorres_options * string) parser =
let
  val option_parser =
    (no_heap_abs_parser ||
     force_heap_abs_parser ||
     ts_rules_parser ||
     ts_force_parser ||
     word_abs_parser ||
     print_stats_parser)
    |> !! (fn xs => K ("autocorres: unknown option " ^ quote (Parse.text (fst xs) |> #1)))

  (* Blank set of options. *)
  val default_opt = {
      no_heap_abs = NONE,
      force_heap_abs = NONE,
      word_abs = NONE,
      ts_rules = NONE,
      ts_force = Symtab.empty,
      print_stats = NONE
    }

  val options_parser = Parse.list option_parser >> (fn opt_fns => fold I opt_fns default_opt)
in
  (* Options *)
  Scan.optional (Parse.$$$ "[" |-- options_parser --| Parse.$$$ "]") default_opt --
  (* Filename *)
  Parse.text
end



(*
 * Worker for the autocorres command.
 *)
fun do_autocorres (opt : autocorres_options) filename thy =
let
  (* Ensure that the filename has already been parsed by the C parser. *)
  val csenv = case CalculateState.get_csenv thy filename of
      NONE => error ("Filename '" ^ filename ^ "' has not been parsed by the C parser yet.")
    | SOME x => x

  (* Enter into the correct context. *)
  val {base = locale_name,...} = OS.Path.splitBaseExt (OS.Path.file filename)
  val lthy = Named_Target.context_cmd (locale_name, Position.none) thy

  (* Fetch basic program information. *)
  val prog_info = ProgramInfo.get_prog_info lthy filename
  val fn_info = FunctionInfo.init_fn_info lthy filename
  val all_functions = Symset.make (Symtab.keys (FunctionInfo.get_functions fn_info))

  (* Process autocorres options. *)

  (* Disallow referring to functions that don't exist. *)
  val funcs_in_options =
        these (#no_heap_abs opt)
        @ these (#force_heap_abs opt)
        @ these (#word_abs opt)
        @ Symtab.keys (#ts_force opt)
  val invalid_functions =
        Symset.subtract all_functions (Symset.make funcs_in_options)
  val _ =
    if Symset.card invalid_functions > 0 then
      error ("autocorres: no such function(s): " ^ commas (Symset.dest invalid_functions))
    else ()

  (* Can only word abstract functions with no callers or callees. *)
  val called_functions =
        flat (map (fn n => FunctionInfo.get_function_callees fn_info n) (Symset.dest all_functions))
  val calling_functions =
        (filter (fn n => length (FunctionInfo.get_function_callees fn_info n) > 0) (Symset.dest all_functions))
  val invalid_word_abs = Symset.inter
      (Symset.make (called_functions @ calling_functions)) (Symset.make (these (#word_abs opt)))
  val _ =
    if not (Symset.is_empty invalid_word_abs) then
      error ("autocorres: Functions that make function calls or are called cannot be word abstracted: "
            ^ commas (Symset.dest invalid_word_abs))
    else ()

  (* Resolve rule names for ts_rules and ts_force. *)
  val ts_force = Symtab.map (K (fn name => Monad_Types.get_monad_type name (Context.Proof lthy)
                                  |> the handle Option => Monad_Types.error_no_such_mt name))
                            (#ts_force opt)
  val ts_rules = Monad_Types.get_ordered_rules (these (#ts_rules opt)) (Context.Proof lthy)

  (* Ensure that we are not both forcing and preventing a function from being heap lifted. *)
  val conflicting_heap_lift_fns =
      Symset.inter (Symset.make (these (#no_heap_abs opt))) (Symset.make (these (#force_heap_abs opt)))
  val _ = if not (Symset.is_empty conflicting_heap_lift_fns) then
            error ("autocorres: Functions are declared as both 'no_heap_abs' and 'force_heap_abs': "
                  ^ commas (Symset.dest conflicting_heap_lift_fns))
          else
            ()

  (* Check that recursive function groups are all lifted to the same monad. *)
  (* FIXME : restore *)
  (*
  val _ = FunctionInfo #recursive_functions prog_info |> Symtab.dest
          |> map (TypeStrengthen.compute_lift_rules ts_rules ts_force o snd)
  *)

  (* (Finished processing options.) *)

  (* Do the translation. *)
  val lthy =
    SimplConv.translate_simpl filename fn_info lthy
    |> (fn (lthy, fn_info) =>
        LocalVarExtract.translate_l2 filename fn_info lthy)
    |> (fn (lthy, fn_info) =>
        HeapLift.system_heap_lift filename fn_info (these (#no_heap_abs opt)) (these (#force_heap_abs opt)) lthy)
    |> (fn (lthy, fn_info) =>
        WordAbstract.word_abstract filename fn_info (these (#word_abs opt)) lthy)
    |> (fn (lthy, fn_info) =>
        TypeStrengthen.type_strengthen ts_rules ts_force filename fn_info lthy)
    |> (fn (lthy, fn_info) =>
        lthy)
in
  (* Exit context. *)
  Local_Theory.exit_global lthy
end

end
