(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Rewrite L2 specifications to use a higher-level ("lifted") heap representation.
 *)

structure HeapLift =
struct

(* Convenience shortcuts. *)
val warning = Utils.ac_warning
val apply_tac = Utils.apply_tac
val the' = Utils.the'

(* Print the current goal state then fail hard. *)
exception ProofFailed of string

fun fail_tac s = (print_tac s THEN (fn _ => Seq.single (raise (ProofFailed s))))

type heap_info = HeapLiftBase.heap_info

(* Return the function for fetching an object of a particular type. *)
fun get_heap_getter (heap_info : heap_info) T =
  case Typtab.lookup (#heap_getters heap_info) T of
    SOME x => Const x
  | NONE => Utils.invalid_typ "heap type for getter" T

(* Return the function for updating an object of a particular type. *)
fun get_heap_setter (heap_info : heap_info) T =
  case Typtab.lookup (#heap_setters heap_info) T of
    SOME x => Const x
  | NONE => Utils.invalid_typ "heap type for setter" T

(* Get the heap type used by a function. *)
fun get_expected_fn_state_type heap_info unliftable_functions fn_name =
  if Symset.contains unliftable_functions fn_name then
    #old_globals_type heap_info
  else
    #globals_type heap_info

(* Get a state translation function for the given function. *)
fun get_expected_st heap_info unliftable_functions fn_name =
  if Symset.contains unliftable_functions fn_name then
    @{mk_term "id :: ?'a => ?'a" ('a)} (#old_globals_type heap_info)
  else
    (#lift_fn_full heap_info)

(* Get the expected type of a function from its name. *)
fun get_expected_hl_fn_type prog_info fn_info (heap_info : HeapLiftBase.heap_info)
    unliftable_functions fn_name =
let
  val fn_def = FunctionInfo.get_function_def fn_info fn_name
  val fn_params_typ = map snd (#args fn_def)
  (* Fill in the measure argument and return type *)
  val globals_typ = get_expected_fn_state_type heap_info unliftable_functions fn_name
  val fn_ret_typ = #return_type fn_def
  val measure_typ = @{typ "nat"}
  val fn_typ = (measure_typ :: fn_params_typ)
                 ---> LocalVarExtract.mk_l2monadT globals_typ fn_ret_typ @{typ unit}
in
  fn_typ
end

(* Get the expected theorem that will be generated about a function. *)
fun get_expected_hl_fn_thm prog_info fn_info (heap_info : HeapLiftBase.heap_info)
    unliftable_functions ctxt fn_name function_free fn_args _ measure_var =
let
  (* Get L2 const *)
  val l2_def = FunctionInfo.get_function_def fn_info fn_name
  val l2_term = betapplys (#const l2_def, measure_var :: fn_args)

  (* Get expected HL const. *)
  val hl_term = betapplys (function_free, measure_var :: fn_args)
in
  @{mk_term "Trueprop (L2Tcorres ?st ?A ?C)" (st, A, C)}
    (get_expected_st heap_info unliftable_functions fn_name, hl_term, l2_term)
end

(* Get arguments passed into the function. *)
fun get_expected_hl_fn_args prog_info fn_info fn_name =
  #args (FunctionInfo.get_function_def fn_info fn_name)

(*
 * Use some heuristics to determine which functions cannot be lifted.
 *
 * For example, we probably can't lift functions that introspect the heap-type
 * data "hrs_htd".
 *)
fun get_unliftable_functions lthy prog_info fn_info =
let
  (* Determine if everything in term "t" appears valid for lifting. *)
  fun can_lift_term t =
  let
    val bad_consts = [@{const_name hrs_htd}, @{const_name hrs_htd_update}, @{const_name ptr_retyp}]
    fun term_contains_const_name c t =
      exists_Const (fn (const_name, _) => c = const_name) t
  in
    not (exists (fn c => term_contains_const_name c t) bad_consts)
  end

  (* Find all functions that are directly unliftable. *)
  val raw_unliftable_functions =
    AutoCorresUtil.map_all fn_info (fn fn_name =>
        (fn_name,
            FunctionInfo.get_function_def fn_info fn_name
            |> #definition
            |> prop_of
            |> Utils.rhs_of
            |> can_lift_term))
    |> filter (not o snd)
    |> map fst
in
  Symset.make raw_unliftable_functions
end

(*
 * Convert a cterm from the format "f a (b n) c" into "((f $ a) $ (b $ n)) $ c".
 *
 * Return a "thm" of the form "old = new".
 *)
fun mk_first_order ctxt ct =
let
  fun expand_conv ct =
    Utils.named_cterm_instantiate
        [("a", Thm.dest_fun ct),("b", Thm.dest_arg ct)] @{lemma "a b == (a $ b)" by simp}
in
  Conv.bottom_conv (K (Conv.try_conv expand_conv)) ctxt ct
end

(* The opposite to "mk_first_order" *)
fun dest_first_order ctxt ct =
  Conv.bottom_conv (K (Conv.try_conv (Conv.rewr_conv
    @{lemma "(op $) == (%a b. a b)" by (rule meta_ext, rule ext, simp)}))) ctxt ct

(*
 * Resolve "base_thm" with "subgoal_thm" in all assumptions it is possible to
 * do so.
 *
 * Return a tuple: (<new thm>, <a change was made>).
 *)
fun greedy_thm_instantiate base_thm subgoal_thm =
let
  val asms = prop_of base_thm |> Logic.strip_assums_hyp
in
  fold (fn (i, asm) => fn (thm, change_made) =>
    if (Term.could_unify (asm, Thm.concl_of subgoal_thm)) then
      (subgoal_thm RSN (i, thm), true) handle (THM _ ) => (thm, change_made)
    else
      (thm, change_made)) (tag_list 1 asms) (base_thm, false)
end

(* Return a list of thm's where "base_thm" has been successfully resolved with
 * one of "subgoal_thms". *)
fun instantiate_against_thms base_thm subgoal_thms =
  map (greedy_thm_instantiate base_thm) subgoal_thms
  |> filter snd
  |> map fst

(*
 * Modify a list of thms to instantiate assumptions where ever possible.
 *)
fun cross_instantiate base_thms subgoal_thm_lists =
let
  fun iterate_base subgoal_thms base_thms =
    map (fn thm => (instantiate_against_thms thm subgoal_thms) @ [thm]) base_thms
    |> List.concat
in
  fold iterate_base subgoal_thm_lists base_thms
end


(* Convert an individual function to use a lift heap state. *)
fun system_heap_lift
    (filename : string)
    (fn_info : FunctionInfo.fn_info)
    (no_heap_abs : string list)
    (force_heap_abs : string list)
    lthy =
let
  val prog_info = ProgramInfo.get_prog_info lthy filename

  (* Create base definitions for the new program, including a new
   * "globals" record with a lifted heap. *)
  val (heap_info, lthy) = HeapLiftBase.setup prog_info fn_info lthy

  (* Extract the abstact term out of a L2Tcorres thm. *)
  fun dest_L2Tcorres_term_abs (_ $ _ $ t $ _ ) = t
  fun get_body_of_thm thm =
      concl_of (Drule.gen_all thm)
      |> HOLogic.dest_Trueprop
      |> dest_L2Tcorres_term_abs

  (* Generate a constant name for our definitions. *)
  fun gen_hl_name x = "hl_" ^ x

  (* Determine which functions cannot be lifted. *)
  val unliftable_functions =
        get_unliftable_functions lthy prog_info fn_info
        |> Symset.union (Symset.make no_heap_abs)
        |> Symset.subtract (Symset.make force_heap_abs)

  (* TODO: explain why each function in the list isn't being lifted. *)
  val _ =
    if Symset.card unliftable_functions > 0 then
      writeln ("HL: Functions cannot be lifted: "
          ^ (commas (Symset.dest unliftable_functions)))
    else
      ()

  (* Declare HL recursion measures. These will be defined during the TypeStrengthen stage. *)
  fun get_measure_typ fn_name =
    let
      val fn_def = FunctionInfo.get_function_def fn_info fn_name
      val fn_params_typ = map snd (#args fn_def)
      val globals = get_expected_fn_state_type heap_info unliftable_functions fn_name
    in
      fn_params_typ ---> (globals --> @{typ "nat"})
    end

  val recursive_functions = FunctionInfo.get_recursive_functions fn_info
  val (measure_consts, lthy) =
      fold Utils.declare_constant
           (map (fn fn_name => (fn_name, "hl_internal_measure'_" ^ fn_name,
                                get_measure_typ fn_name)) recursive_functions)
           (Symtab.empty, lthy)

  (* Translate L2 measures to the lifted heap representation. *)
  fun l2_hl_measure_def fn_name =
    let
      val l2_m = Utils.get_term lthy ("l2_internal_measure'_" ^ fn_name)
      val hl_m = the' ("Looking up hl_internal_measure'_" ^ fn_name)
                      (Symtab.lookup measure_consts fn_name)
      val fn_def = FunctionInfo.get_function_def fn_info fn_name
      val fn_params_vars = map Free (#args fn_def)
      val state_var_name = if member (op =) (map fst (#args fn_def)) "s"
                              then "s'" else "s"
      val state_var = Free (state_var_name, #old_globals_type heap_info)
      val new_state = if Symset.contains unliftable_functions fn_name
                         then state_var else #lift_fn_full heap_info $ state_var

      val hl_m_call = betapplys (hl_m, fn_params_vars @ [new_state])
                      |> fold lambda (rev (fn_params_vars @ [state_var]))
      val def_term = @{mk_term "?m == ?def" (m, def)} (l2_m, hl_m_call)
    in
      def_term
    end

  val (l2_measure_defs0, lthy) =
      fold Utils.define_constant
           (map (fn fn_name => (fn_name,
                                "l2_internal_measure'_" ^ fn_name ^ "_def",
                                "l2_internal_measure'_" ^ fn_name,
                                l2_hl_measure_def fn_name))
                recursive_functions)
           (Symtab.empty, lthy)

  (* We also want to prove "hl_measure_valid" correspondence for each constant. *)
  fun hl_measure_valid_thm (fn_name, measure_def) (l2_measure_defs, lthy) =
    let
      val l2_m = Utils.get_term lthy ("l2_internal_measure'_" ^ fn_name)
      val hl_m = the' ("Looking up hl_internal_measure'_" ^ fn_name)
                      (Symtab.lookup measure_consts fn_name)
      val fn_def = FunctionInfo.get_function_def fn_info fn_name
      val goal = let
          val fn_params_vars = map Free (#args fn_def)
          val l2_apply = betapplys (l2_m, fn_params_vars)
          val hl_apply = betapplys (hl_m, fn_params_vars)
          val st = get_expected_st heap_info unliftable_functions fn_name
          val stmt = (@{const "Trueprop"} $
                      @{mk_term "hl_measure_valid ?st ?hl_app ?l2_app"
                        (st, hl_app, l2_app)} (st, hl_apply, l2_apply))
                     |> fold Logic.all (rev (* aesthetic *) fn_params_vars)
      in stmt end

      val measure_valid_thm =
        Goal.prove lthy (map fst (#args fn_def)) [] goal
          (fn _ => simp_tac (HOL_ss addsimps [@{thm id_def}, @{thm hl_measure_valid_def}, measure_def]) 1)
      val measure_corres_thms =
          [ measure_valid_thm RS @{thm L2Tcorres_measured_call} ] @
          (if Symset.contains unliftable_functions fn_name
           then [ measure_valid_thm RS @{thm L2Tcorres_exec_concrete_measured} ]
           else [ measure_valid_thm RS @{thm L2Tcorres_exec_abstract_measured} ])
    in
      (Symtab.update_new (fn_name, (measure_def, measure_corres_thms)) l2_measure_defs, lthy)
    end

  val (l2_measure_defs, lthy) =
      fold hl_measure_valid_thm (Symtab.dest l2_measure_defs0) (Symtab.empty, lthy)

  (* Generate "valid_typ_heap" predicates for each heap type we have. *)
  fun mk_valid_typ_heap_thm typ =
    @{mk_term "Trueprop (valid_typ_heap ?st ?getter (%val. ?setter (%_. val)) ?t_hrs ?t_hrs_update)" (st, getter, setter, t_hrs, t_hrs_update)}
      (#lift_fn_full heap_info,
          get_heap_getter heap_info typ,
          get_heap_setter heap_info typ,
          #t_hrs_getter prog_info,
          #t_hrs_setter prog_info)
    |> (fn prop => Goal.prove_future lthy [] [] prop
         (K ((rtac @{thm valid_typ_heapI} 1)
         THEN (REPEAT (
          (CHANGED (asm_full_simp_tac
            ((simpset_of lthy) addsimps [#lift_fn_thm heap_info] @ @{thms typ_simple_heap_simps}) 1))
          ORELSE (rtac @{thm ext} 1))))))

  (* Make thms for all types. *)
  val heap_types = (#heap_getters heap_info |> Typtab.dest |> map fst)
  val valid_typ_heap_thms = map mk_valid_typ_heap_thm heap_types

  (* Generate "valid_struct_field" for each field of each struct. *)
  fun mk_valid_struct_field_thm typ (field_info : HeapLiftBase.field_info) =
    @{mk_term "Trueprop (valid_struct_field ?st [?fname] ?fgetter (%v. ?fsetter (%_. v)) ?getter ?setter ?t_hrs ?t_hrs_update)" (st, fname, fgetter, fsetter, getter, setter, t_hrs, t_hrs_update) }
      (#lift_fn_full heap_info,
          Utils.encode_isa_string (#name field_info),
          #getter field_info,
          #setter field_info,
          get_heap_getter heap_info typ,
          get_heap_setter heap_info typ,
          #t_hrs_getter prog_info,
          #t_hrs_setter prog_info)
    |> (fn prop => Goal.prove_future lthy [] [] prop
         (K ((rtac @{thm valid_struct_fieldI} 1)
         THEN (REPEAT (
          (CHANGED (asm_full_simp_tac
            ((simpset_of lthy) addsimps [#lift_fn_thm heap_info] @ @{thms typ_simple_heap_simps}) 1))
          ORELSE (rtac @{thm ext} 1))))))

  (* Make thms for all fields of structs in our heap. *)
  fun valid_struct_abs_thms T =
    case (Typtab.lookup (#struct_types heap_info) T) of
      NONE => []
    | SOME struct_info =>
        map (mk_valid_struct_field_thm T) (#field_info struct_info)
  val valid_field_thms =
    map valid_struct_abs_thms heap_types |> List.concat

  (* Generate conversions from globals embedded directly in the "globals" and
   * "new_globals" record. *)
  fun mk_valid_globals_field_thm name =
    @{mk_term "Trueprop (valid_globals_field ?st ?old_get ?old_set ?new_get ?new_set)" (st, old_get, old_set, new_get, new_set)}
      (#lift_fn_full heap_info,
        Symtab.lookup (#global_field_getters heap_info) name |> the |> fst,
        Symtab.lookup (#global_field_setters heap_info) name |> the |> fst,
        Symtab.lookup (#global_field_getters heap_info) name |> the |> snd,
        Symtab.lookup (#global_field_setters heap_info) name |> the |> snd)
    |> (fn prop => Goal.prove_future lthy [] [] prop
         (K (asm_full_simp_tac ((simpset_of lthy)
           addsimps [#lift_fn_thm heap_info, @{thm valid_globals_field_def}]) 1)))
  val valid_global_field_thms = map (#1 #> mk_valid_globals_field_thm) (#global_fields heap_info)

  (*
   * Fetch rules from the theory, instantiating any rule with
   * "valid_globals_field", "valid_typ_heap" etc with the predicates generated
   * above.
   *)
  val rules =
      cross_instantiate (HeapAbsThms.get lthy) [
        valid_typ_heap_thms,
        valid_field_thms,
        valid_global_field_thms]
    |> filter (prop_of #> exists_subterm (fn x => case x of Const (@{const_name "valid_globals_field"}, _) => true | _ => false) #> not)
    |> filter (prop_of #> exists_subterm (fn x => case x of Const (@{const_name "valid_struct_field"}, _) => true | _ => false) #> not)
    |> filter (prop_of #> exists_subterm (fn x => case x of Const (@{const_name "valid_typ_heap"}, _) => true | _ => false) #> not)

  val l2_measures_valid_thms = Symtab.dest l2_measure_defs |> map snd |> map snd |> List.concat

  (* Instantiate recursive L2_call rule *)
  val measure_term = cterm_of (Proof_Context.theory_of lthy)
                              (@{const "recguard_dec"} $ Free ("measure'", @{typ "nat"}))
  val L2Tcorres_rec_call_inst = @{thm L2Tcorres_rec_call} WHERE [("m", measure_term)]

  val rules = rules @ [L2Tcorres_rec_call_inst] @ l2_measures_valid_thms

  (* Convert to new heap format. *)
  fun convert ctxt fn_name callee_terms measure_var fn_args =
  let
    val thy = Proof_Context.theory_of ctxt

    (* Fetch the function definition. *)
    val fn_def = FunctionInfo.get_function_def fn_info fn_name
    val l2_body_def =
        #definition fn_def
        (* Instantiate the arguments. *)
        |> Utils.inst_args (map (cterm_of thy) (measure_var :: fn_args))
    val l2_body = Utils.rhs_of (prop_of l2_body_def)

    (* Get L2 body definition with function arguments. *)
    val l2_term = betapplys (#const fn_def, measure_var :: fn_args)

    (* Get our state translation function. *)
    val st = get_expected_st heap_info unliftable_functions fn_name

    (* Generate a schematic goal. *)
    val goal = @{mk_term "Trueprop (L2Tcorres ?st ?A ?C)" (st, C)}
        (st, l2_term)
        |> cterm_of thy
        |> Goal.init
        |> Utils.apply_tac "unfold RHS" (EqSubst.eqsubst_tac ctxt [0] [l2_body_def] 1)

    val rules = rules @ (map (snd #> #3) (Symtab.dest callee_terms))
    val fo_rules = HeapAbsFOThms.get ctxt

    (* Apply a conversion to the concrete side of the given L2T term. *)
    fun l2t_conc_body_conv conv =
      Conv.params_conv (~1) (K (Conv.arg_conv (Utils.nth_arg_conv 4 conv)))

    (* Standard tactics. *)
    fun rtac_all r n = (APPEND_LIST (map (fn thm => rtac thm n) r))

    (* Convert the concrete side of the given L2T term to/from first-order form. *)
    val l2t_to_fo_tac = CONVERSION (Drule.beta_eta_conversion then_conv l2t_conc_body_conv (mk_first_order ctxt) ctxt)
    val l2t_from_fo_tac = CONVERSION (l2t_conc_body_conv (dest_first_order ctxt then_conv Drule.beta_eta_conversion) ctxt)
    val fo_tac = ((l2t_to_fo_tac THEN' rtac_all fo_rules) THEN_ALL_NEW l2t_from_fo_tac) 1

    (* If debugging enabled, print goal states when we backtrack. *)
    val error_printed = ref 0
    fun print_n_tac printed_ref t =
      (if true orelse (!printed_ref >= get_print_depth ()) then
        all_tac
      else
        (printed_ref := (!printed_ref) + 1; print_tac "Error solving subgoal")) t

    (*
     * Recursively solve subgoals.
     *
     * We allow backtracking in order to solve a particular subgoal, but once a
     * subgoal is completed we don't ever try to solve it in a different way.
     *
     * This allows us to try different approaches to solving subgoals without
     * leading to exponential explosion (of many different combinations of
     * "good solutions") once we hit an unsolvable subgoal.
     *)
    fun solve_tac n =
      DETERM ((SOLVES (
          ((K (CHANGED ((rtac_all rules 1) APPEND (fo_tac)))) THEN_ALL_NEW solve_tac) 1))
            ORELSE (print_n_tac error_printed THEN no_tac))
    val thm = Utils.apply_tac "Solve goal" (solve_tac 1) goal
      |> Goal.finish ctxt

    (* Gather statistics. *)
    val _ = Statistics.gather ctxt "HL" fn_name
        (Thm.prop_of thm |> HOLogic.dest_Trueprop |> (fn t => Utils.term_nth_arg t 2))

    (* Apply peephole optimisations to the theorem. *)
    val _ = writeln ("Simpifying (HL) " ^ fn_name)
    val thm = L2Opt.cleanup_thm ctxt thm false 2

    (* Gather post-simplification statistics. *)
    val _ = Statistics.gather ctxt "HLsimp" fn_name
        (Thm.prop_of thm |> HOLogic.dest_Trueprop |> (fn t => Utils.term_nth_arg t 2))
  in
    (get_body_of_thm thm, Drule.gen_all thm)
  end

  (* Update function information. *)
  fun update_function_defs lthy fn_def =
    FunctionInfo.fn_def_update_const (Utils.get_term lthy (gen_hl_name (#name fn_def))) fn_def
    |> FunctionInfo.fn_def_update_definition (
        (the (AutoCorresData.get_def (Proof_Context.theory_of lthy)
            filename "HLdef" (#name fn_def))))
in
  AutoCorresUtil.do_translation_phase
    "HL" filename fn_info
    (get_expected_hl_fn_type prog_info fn_info heap_info unliftable_functions)
    (get_expected_hl_fn_thm prog_info fn_info heap_info unliftable_functions)
    (get_expected_hl_fn_args prog_info fn_info)
    gen_hl_name
    convert
    update_function_defs
    @{thm L2Tcorres_recguard_0}
    lthy
end

end
