(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Utiltites for working with HOL records.
 *)
structure RecordUtils =
struct

(* Fetch information about a record from the record package. *)
fun get_record_info thy T =
  Record.dest_recTs T
  |> hd
  |> fst
  |> Long_Name.explode
  |> split_last
  |> fst
  |> Long_Name.implode
  |> Record.the_info thy

(*
 * Get the constructor for the given record type; i.e., the constant
 * which is used to construct records of this type.
 *
 * We instantiate the "more" component of the record type to unit.
 *
 * HACK: This doesn't seem to be the best way of doing this, because we
 * rely on the ordering of the "#defs" field of the record_info
 * structure.
 *)
fun get_record_constructor thy T =
  get_record_info thy T |> #ext_def |> prop_of |> Utils.lhs_of |> head_of
  |> Term_Subst.instantiate ([((("'z", 0), @{sort "type"}), @{typ unit})], [])

(*
 * Given a record "bar" (assumed to have fields "a", "b", "c" and a
 * definition of the form:
 *
 *   foo x y s == bar_ext (f x y s) (f' x y s) (f'' x y s)
 *
 * generate simps of the form:
 *
 *   a (foo x y s) = f x y s
 *   b (foo x y s) = f' x y s
 *   c (foo x y s) = f'' x y s
 *
 * This allows certain simplifications to occur on the definition
 * "foo" without having to expand it.
 *)
fun generate_ext_simps name def_thm lthy =
let
  val thy = Proof_Context.theory_of lthy

  (* Determine record type. *)
  val T = prop_of def_thm |> Utils.rhs_of |> head_of |> fastype_of |> body_type

  (* Fetch components that make up the new record. *)
  val components = prop_of def_thm |> Utils.rhs_of |> strip_comb |> snd

  (* Fetch fields of the record type T. *)
  val fields = let
    val (a, b) = Record.get_recT_fields thy T
  in
    a @ [b]
  end

  (* Generate theorems. *)
  fun gen_simp_proof ((field_name, field_type), defn) =
  let
    val ss = simpset_of lthy
  in
      HOLogic.mk_Trueprop (
        HOLogic.mk_eq (
          Const (field_name, T --> field_type) $ (prop_of def_thm |> Utils.lhs_of), defn))
      |> cterm_of thy
      |> Goal.init
      |> Utils.apply_tac "solve simple record proof"
          (simp_tac (ss addsimps [def_thm]) 1)
      |> Goal.finish lthy
  end
  val thms = map gen_simp_proof (fields ~~ components)

  (* Define the theorem. *)
  val lthy = Utils.define_lemmas name thms lthy |> snd

  (* Add to new rules to the simpset. *)
  val lthy = Local_Theory.map_contexts
    (fn _ => Context.proof_map (Simplifier.map_ss (fn x => x addsimps thms))) lthy
in
  (thms, lthy)
end

(* Get a record setter from its getter name and type. *)
fun get_record_getter recT (name, T) =
  Const (name, recT --> T)
fun get_record_setter recT (name, T) =
  Const (name ^ "_update", (T --> T) --> recT --> recT)

end
