(*
 * Copyright (C) 2013, National ICT Australia Limited. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *  * Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * The name of National ICT Australia Limited nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 * IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *)

(*
 * Rewrite L2 specifications to use "nat" and "int" data-types instead of
 * "word" data types. The former tend to be easier to reason about.
 *)

structure WordAbstract =
struct

(* Convenience shortcuts. *)
val warning = Utils.ac_warning
val apply_tac = Utils.apply_tac
val the' = Utils.the'

(* Get a function's name from its heap-lifted name. *)
fun guess_l2_fn_name n = (unprefix "l2_" (Long_Name.base_name n))

(* Get abstract version of a HOL type. *)
fun get_abs_type T =
  if T = @{typ word32} then
    @{typ nat}
  else if T = @{typ sword32} then
    @{typ int}
  else T

(* Get abstraction function for a HOL type. *)
fun get_abs_fn T =
  if T = @{typ word32} then
    @{term "unat :: word32 => nat"}
  else if T = @{typ sword32} then
    @{term "sint :: sword32 => int"}
  else @{mk_term "%x :: ?'a. x" ('a)} T

fun get_abs_inv_fn t =
  if fastype_of t = @{typ nat} then
    @{term "of_nat :: nat => word32"} $ t
  else if fastype_of t = @{typ int} then
    @{term "of_int :: int => sword32"} $ t
  else t

(* Get the expected type of a function from its name. *)
fun get_expected_fn_type prog_info fn_info fn_name =
let
  val fn_def = FunctionInfo.get_function_def fn_info fn_name
  val fn_params_typ = map (get_abs_type o snd) (#args fn_def)
  val fn_ret_typ = get_abs_type (#return_type fn_def)
  val globals_typ = LocalVarExtract.dest_l2monad_T (fastype_of (#const fn_def)) |> snd |> #1
  val measure_typ = @{typ "nat"}
in
  (measure_typ :: fn_params_typ)
      ---> LocalVarExtract.mk_l2monadT globals_typ fn_ret_typ @{typ unit}
end

(* Get the expected theorem that will be generated about a function. *)
fun get_expected_fn_thm prog_info fn_info ctxt fn_name
    function_free fn_args _ measure_var =
let
  val old_def = FunctionInfo.get_function_def fn_info fn_name
  val old_term = betapplys (#const old_def, measure_var :: map get_abs_inv_fn fn_args)
  val new_term = betapplys (function_free, measure_var :: fn_args)
in
  @{mk_term "Trueprop (corresTA ?rt (%x. x) ?A ?C)" (rt, A, C)}
      (get_abs_fn (#return_type old_def), new_term, old_term)
end

(* Get arguments passed into the function. *)
fun get_expected_fn_args prog_info fn_info fn_name =
  map (apsnd get_abs_type) (#args (FunctionInfo.get_function_def fn_info fn_name))

(* Convert an individual function to use a lift heap state. *)
fun word_abstract filename fn_info target_functions lthy =
let
  val prog_info = ProgramInfo.get_prog_info lthy filename

  (* Generate a constant name for our definitions. *)
  fun gen_wa_name x = "wa_" ^ x

  (*
   * Fetch rules from the theory, instantiating any rule with
   * "valid_globals_field", "valid_typ_heap" etc with the predicates generated
   * above.
   *)
  val rules = WordAbsThms.get lthy

  (* Abstract functions. *)
  fun convert ctxt fn_name callee_terms measure_var fn_args =
  let
    val thy = Proof_Context.theory_of ctxt
    val old_fn = FunctionInfo.get_function_def fn_info fn_name

    (* Construct arguments for the concrete body. *)
    val conc_args = map get_abs_inv_fn fn_args

    (* Fetch the function definition, and instantiate its arguments. *)
    val old_body_def =
        #definition old_fn
        (* Instantiate the arguments. *)
        |> Utils.inst_args (map (cterm_of thy) (measure_var :: conc_args))
    val old_body = Utils.rhs_of (prop_of old_body_def)

    (* Get old body definition with function arguments. *)
    val old_term = betapplys (#const old_fn, measure_var :: conc_args)

    (* Generate a schematic goal. *)
    val goal = @{mk_term "Trueprop (corresTA ?ra (%x. x) ?A ?C)" (ra, C)}
            (get_abs_fn (#return_type old_fn), old_term)
        |> cterm_of thy
        |> Goal.init
        |> Utils.apply_tac "unfold RHS" (EqSubst.eqsubst_tac ctxt [0] [old_body_def] 1)

    (* Fetch rules tagged "word_abs", plus rules for our callees. *)
    val rules = rules @ (map (snd #> #3) (Symtab.dest callee_terms))

    (* Standard tactics. *)
    fun rtac_all r n = (APPEND_LIST (map (fn thm => rtac thm n) r))

    (* If debugging enabled, print goal states when we backtrack. *)
    val error_printed = ref 0
    fun print_n_tac printed_ref t =
      (if not (!Runtime.debug) orelse (!printed_ref >= get_print_depth ()) then
        all_tac
      else
        (printed_ref := (!printed_ref) + 1; print_tac "Error solving subgoal")) t

    (*
     * Recursively solve subgoals.
     *
     * We allow backtracking in order to solve a particular subgoal, but once a
     * subgoal is completed we don't ever try to solve it in a different way.
     *
     * This allows us to try different approaches to solving subgoals without
     * leading to exponential explosion (of many different combinations of
     * "good solutions") once we hit an unsolvable subgoal.
     *)
    fun solve_tac n =
        DETERM (( SOLVES (
            ((K (CHANGED ((rtac_all rules 1) APPEND (no_tac)))) THEN_ALL_NEW solve_tac) 1))
            ORELSE (print_n_tac error_printed THEN no_tac))
    val thm = Utils.apply_tac "Solve goal" (solve_tac 1) goal
      |> Goal.finish ctxt

    (* Apply peephole optimisations to the theorem. *)
    val _ = writeln ("Simpifying (WA) " ^ fn_name)
    val thm = L2Opt.cleanup_thm ctxt thm false 3

    (* Extract the abstact term out of a L2Tcorres thm. *)
    fun dest_corresWA_term_abs (_ $ _ $ _ $ t $ _ ) = t
    fun get_body_of_thm thm =
        concl_of (Drule.gen_all thm)
        |> HOLogic.dest_Trueprop
        |> dest_corresWA_term_abs

  in
    (get_body_of_thm thm, Drule.gen_all thm)
  end

  (* Update function information. *)
  fun update_function_defs lthy fn_def =
    FunctionInfo.fn_def_update_const (Utils.get_term lthy (gen_wa_name (#name fn_def))) fn_def
    |> FunctionInfo.fn_def_update_definition (
        (the (AutoCorresData.get_def (Proof_Context.theory_of lthy)
            filename "WAdef" (#name fn_def))))
    |> FunctionInfo.fn_def_update_return_type (get_abs_type (#return_type fn_def))
    |> FunctionInfo.fn_def_update_args (map (apsnd get_abs_type) (#args fn_def))
in
  AutoCorresUtil.do_partial_translation_phase
    target_functions
    "WA" filename fn_info
    (get_expected_fn_type prog_info fn_info)
    (get_expected_fn_thm prog_info fn_info)
    (get_expected_fn_args prog_info fn_info)
    gen_wa_name
    convert
    update_function_defs
    @{thm L2Tcorres_recguard_0}
    lthy
end

end
